.\" Man page for prog84, a 16C84 programmer driver
.\" 
.\" this manualpage may be copied over to /usr/local/man/man1/prog84.1
.\" 
.\" 

.TH prog84 1 "14th. Februar 2002"
.SH NAME
prog84 \- program a Microchip Technology eeprom like 24C16,
or a PIC through the parallel, serial or USB port.
PIC12C5xx, PIC16C84/16F84  and many other PIC's from the
16Cxxx and 16F87x series are supported by prog84.

.SH SYNOPSIS
\fBprog84\fP [\fB-abhvzTQFL\fP] [\fB-P\fP \fIspeed\fP] [\fB-i\fP \fIn1,n2,n3,n4\fP] [\fB-c\fP \fInn\fP] [\fB-C\fP \fIxxx\fP] [[\fB-p\fP|\fB-x\fP|\fB-y\fP] \fIfile\fP]
.SH DESCRIPTION
The
.B prog84
driver attempts to program a 16Cxx and 16Fxxx EEPROM/FLASH microcontroller
series and 12C5xx EPROM microconrollers
in serial mode through a minimal driver circuit connected to a IBM-PC style
parallel port, serial port, or USB parallelport.
It is designed to work easily with any such circuit
by using the correct configuration file. See the HARDWARE section
for information on the external circuitry required and how to specify
it to

.B Warning!
The 12C5xx support is in very alpha stage, and has some limitations.
See 12C5xx section for further details.

.BR prog84 .
.PP
By default, 
.B prog84
does nothing when invoked. These command line options control what happens,
in the order they are specified on the command line.
.PP
.TP
\fB-h\fP , \fB--help\fP
Emits a brief help message listing the acceptable command line
parameters.

.TP
\fB-x\fP \fIfilename\fP , \fB--intel8\fP=\fIfilename\fP
Program the PIC with the data contained in the 
8-bit Intel 
.BR hex -format
file
.IR filename .

.TP
\fB-y\fP \fIfilename\fP , \fB--intel16\fP=\fIfilename\fP
Program the PIC with the data contained in the 
16-bit Intel 
.BR hex -format
file
.IR filename .

.TP
\fB-p\fP \fIfilename\fP , \fB--pictools\fP=\fIfilename\fP 
Program the PIC with the data contained in the 
.BR pictools -format
file
.IR filename .


.TP
\fB-i\fP \fIn1,n2,n3,n4\fP , \fB--id\fP=\fIn1,n2,n3,n4\fP
Program the chip's ID words with the specified values. Each value is
14 bits specified in C notation: decimal, or 
.B 0
followed by an octal value, or
.B 0x
followed by a value in hexadecimal.


.TP
\fB-O\fP \fInn\fP  , \f--osc-val\fP=\fInn\fP
8 bit oscillator calibration value for UV erasable 12c5xx chips.
The option always override the value in hex file.
Number may be decimal, octal or hexadecimal as above.


.TP
\fB-c\fP \fInn\fP  , \fB--fuses-val\fP=\fInn\fP  
Program the fuses with the specified value. The value may use the
same formats as descrived above for \fB-i\fP.
If this option is located before \fB-x\fP or \fB-p\fP, this value
will override any value in the hex or pictool file.


.TP
\fB-C\fP \fIxxx\fP , \fB--fuses\fP=\fIxxx\fP
Program the specified fuses. 
\fIxxx\fP may be any of following:
   \fIC\fP : code protect on
.br
   \fIW\fP : watchdog timer  on
.br
   \fIU\fP : power up timer  on
.br

Only one of following oscillator settings should be set:
   \fIL\fP : LP    \fIX\fP : XT    \fIH\fP : HS    \fIR\fP : RC  

If this option is located before \fB-x\fP or \fB-p\fP, this value
will override any value in the hex or pictool file.

Example for 16f84:
   prog84 -zCux -x myfile.hex

Example for 16c84:
   prog84 -zCx -x myfile.hex


.TP
\fB-P\fP \fIspeed\fP , \fB--serial\fP=\fIspeed\fP
When using serial PIC programmer, apply power and run serial port 
at given speed. Usually default settings works.

.TP
\fB-Q\fP , \fB--power-off\fP
Remove power from the device.

.TP
\fB-z\fP , \fB--clear\fP
Erase the PIC or EEPROM. 

.TP
\fB-Z\fP\fINN\fP , \fB--eeprom\fP=\fINN\fP
Only works with  EEPROM's (24cXX). Every byte of the EEPROM is programmed 
with the specified value \fINN\fP.

.PP
The following options modify the behavior of the programming options.
.TP
\fB-a\fP , \fB--verify\fP
After writing any location, read it back to verify that it was correctly
programmed.

.TP
\fB-W\fP . \fB--write-diff\fP
(default)
Before writing any location, read it to see if it already contains the
desired value. If so, the actual programming step is skipped. It's probably
best to specify both the
.B -a
and 
.B -W
options when programming a microcontroller.
.TP
\fB-F\fP , \fB--write-all\fP
Force programming even if value is already correct.

.TP
\fB-T\fP \fIchiptype\fP , \fB--chip\fP=\fIchiptype\fP
Select type of chip. Currently the following values
can be used:

"a", "auto", "h", "help", 2408, 2416, 2465, 1684, and 
values for the series 16C6x/7x/923/924 and 16F87x. 12508 and 12509
values stands for 12C5xx series.

Use 
.B -T \fIhelp\fP
or
.B -T \fIh\fP
to get the full list.

.B -T \fIa\fP
or
.B -T \fIauto\fP
will try to autodetect eeprom, and if none is detected an 16x84 is
chosen.
Selecting an eeprom will on BR875 and equivalent programmers autoselect
the eeprom (since BR875 does not have any switches).

.TP
\fB-v\fP , \fB--verbose\fP
Be more verbose. This option can be specified multiple times to increase
the amount of output produced. One
.B -v
elicits a terse status display, showing which locations are being
programmed and whether they verify correctly. Two will show
the results of the initial hardware test and dump the information read
from the configuration file. Three will produce a line of output
describing every bit clocked into the PIC and the bit read back
(probably more information than is useful on a regular basis).

.TP
\fB-L\fP , \fB--list-chips\fP
Prints a list of all supported devices. The list also shows the type of
device (\fIPIC\fP or \fIEEPROM\fP) and the amount of memory available
for programming.

.TP
\fB-V\fP , \fB--version\fP
Print version information.

.PP


.SH HADRWARE
The circuitry connected to the parallel port must have the ability to
control the power, MCLR, clock (RB6), and data (RB7) pins of the
16C84, as well as the ability to read back the state of the data (RB7)
pin.
.PP
Information on the driver circuitry is read from the file
.B prog84rc
in the working directory when 
.B prog84
is started. 
If this file does not exist, then
.B ~/.prog84rc
is tried.
This file contains pin descriptions, one per line. Comments
can be started with a hash mark and continue to the end of the line.
Blank lines are ignored. 

.PP
The following keywords are available:

.TP
.BI port  \ =  \ serial|parallel|usb  
This line must be present to select the hardware.

.TP
.BI no_power_C4C8 \ = \ 0|1
Default value is 
.B 0
if the keyword is omitted. Must be set to
.B 1
when using the BR875 pic-programmer.

.TP
.BI loop \ = \ <N>
This keyword is only used in the DOS version. The value
.B N
must be computed by the DOS-program
.I dosinit.exe .

.TP
.BR base \ = \ 0xNNN
If parallel or serial  port is at a base address other than 0x378 hex, 
you can specify the base address with a line of the form "base = 0x3bc".


Each data line must contain a pin name, a
colon or equals sign, and a description of the parallel-port bit 
used for that pin. 
.PP
A pin description can be \fBd\fP\fIn\fP to specify data bit \fIn\fP;
\fBc\fP\fIn\fP to specify control bit \fIn\fP, or
\fBs\fP\fIn\fP to specify status bit \fIn\fP.
The bits are numbered starting with the LSB as 0, in the order
they are found in the PC's I/O space. Any pin can be preceded
by a tilde (~) or exclamation point (!) to reverse the sense of
the bit, \fIe.g.\fP if the driver circuit uses inverting buffers.
Additionally, some symbolic names are understood: \fBbusy\fP refers
to status bit 7; \fBack\fP to status bit 6; and \fBerror\fP, bit 3.
.PP
The programmer requires the following bits to be specified:
.TP
\fBpower\fP
Set to 1 when the programmer wants the PIC's Vdd at +5 volts.
.TP
\fBmclr\fP
Set to 1 when the programmer wants the PIC's /MCLR input set to
the programming voltage (+13 volts). That's the Vpp voltage.
.TP
\fBclock\fP
Used to clock data into the PIC (connect to PIC pin 12)
.TP
\fBdata\fP
Used to write data and commands to the PIC (connect to PIC pin 13)
.TP
\fBdata_f\fP
A return line to read the state of the data pin
.TP
\fBclock_f\fP
Optional; if specified, this should read the state of the clock pin
at the PIC, and is used in the test sequence. If verbose mode is
selected a warning will be issued reporting that 
.B clock_f
is emulated. 
.TP
\fBvpp2\fP
Optional; if specified, this line is ever set to 0. This line is a
secondary Vpp pin and isn't currently used, but is needed to avoid
damaging the chip when using the PROPIC circuit.
.PP
Even if your hardware allows you to read from bits other than the
status bits (\fIe.g.\fP a bidirectional parallel port), 
.B prog84
is not smart enough to actually try to read from them.
The data line is assumed to be open-collector; it will be driven
high (1) when the driver wishes to read data back from the PIC.
.SH BUGS
The
.B pictools
file format is pretty esoteric, being produced by only one known 
assembler. 
.B prog84
was therefor extended to read the common Intel Hex format, both
the 8-bit and 16-bit hexfiles can be read. Currently the
file checksum information in the file is ignored.
.PP
The program should be smart enough to use the parallel programming mode
if enough pins are available.
.PP
Because of realtime timing requirements which are hard to meet in user
mode on a Unixlike operating system, 
.B prog84
cannot program the other members of the PIC family.
.PP
The configuration file,
.BR prog84rc ,
is always read from the current working directory, and has a somewhat
restricted syntax. In particular, it would be useful to specify bits
in terms of their positions on the DB-25 connector.

.SH 12C5xx

12C5xx extension is tested on JDM serial programmer only.
.BR dump84
and
.BR prog84
programs work. No other utilities are tested.
Due to differences between PIC16 and PIC12 memory organization and
configuration word, the following limitations should be applied:

\fB-C\fP option should not be used. Configuration word may be programmed
via the input file made by your assembler, or by \fB-c\fP option.

Input file can be Intel hexa only (\fB-x\fP, \fB-y\fP). Do not use
.B pictools
format (\fB-p\fP).

EPROM based devices cannot be cleared eletrically.
\fB-z\fP option has no effect.

.SH SEE ALSO
The
.B pictools
assembler is written and maintained by Ian King, 
iek@advanced-robotics-research-centre.salford.ac.uk
and is available via anonymous FTP from ftp.funet.fi
in the /pub/microprocs/pic directory.
.PP
The
.B picasm
asembler is written by Timo Rossi, and is available at
at http://www.iki.fi/trossi/pic/ .

.PP
This code is based on the specifications in the Microchip Technology
1994 Data Book. At http://www.microchip.com/ many specifications
can be downloaded.
.SH AUTHOR
Written and copyright by Wim Lewis, 
wiml@netcom.com, wiml@omnigroup.com, wiml@hhhh.org.
You may use, modify, and redistribute this code as long as you retain
the attributions and do not incorporate it into a commercial product.

Updated by Frank Damgaard, frda@post3.tele.dk, to conform with
prog84 version 3.0. No manual page yet for dump84 and thru84,
use the 
.BR -h
option 
for an list of available options.

12c5xx extension is added by Gabor Kiss, kissg@sztaki.hu.
The work was inspired by and some ideas was stolen from
Jens Dyekjaer Madsen's original Pascal program driving his
excellent PIC programmer.
